#include <assert.h>
#include <limits.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fio.h"
#include "gfx.h"
#include "handles.h"
#include "renderer_api.h"
#include "runtime.h"

typedef enum {
    VY_STMT_FORM_VALUE,
    VY_STMT_FORM_LIST,
} vy_stmt_form;

typedef struct {
    unsigned int first;
    unsigned int count;
} vy_parsed_stmt_list;

typedef struct {
    vy_stmt_form form;
    vy_text_span attribute;
    union {
        vy_text_span value;
        unsigned int list_index;
    };
    /* For lists */
    unsigned int next;
} vy_parsed_stmt;

typedef struct {
    const char *file;
    const char *text;
    size_t at;
    size_t length;
    int line;

    vy_parsed_stmt *statements;
    unsigned int statement_count;
    unsigned int statement_capacity;

    vy_parsed_stmt_list *statement_lists;
    unsigned int statement_list_count;
    unsigned int statement_list_capacity;
} vy_parse_state;

static bool IsAllowedChar(char c) {
    return (c >= '0' && c <= '9') || (c >= 'a' && c <= 'z') ||
           (c >= 'A' && c <= 'Z') || (c == '.') || (c == '_') || (c == '/');
}

static bool IsWhitespace(char c) {
    return c == ' ' || c == '\t' || c == '\n' || c == '\r';
}

static void SkipWhitespace(vy_parse_state *state) {
    while (state->at < state->length && IsWhitespace(state->text[state->at])) {
        if (state->text[state->at] == '\n')
            ++state->line;
        ++state->at;
    }
}

static bool ParseAttribute(vy_parse_state *state, vy_text_span *_name) {
    vy_text_span name;
    name.start  = &state->text[state->at];
    name.length = 0;
    while (state->at < state->length && !IsWhitespace(state->text[state->at])) {
        if (IsAllowedChar(state->text[state->at])) {
            ++state->at;
            ++name.length;
        } else {
            vyReportError("GFX",
                          "%s:%d Unexpected character %c",
                          state->file,
                          state->line,
                          state->text[state->at]);
            return false;
        }
    }
    *_name = name;
    return true;
}

static bool ParseValue(vy_parse_state *state, vy_text_span *_value) {
    vy_text_span value;
    value.start  = &state->text[state->at];
    value.length = 0;
    while (state->at < state->length && !IsWhitespace(state->text[state->at]) &&
           state->text[state->at] != ';') {
        if (IsAllowedChar(state->text[state->at])) {
            ++state->at;
            ++value.length;
        } else {
            vyReportError("GFX",
                          "%s:%d Unexpected character %c",
                          state->file,
                          state->line,
                          state->text[state->at]);
            return false;
        }
    }
    *_value = value;
    return true;
}

static bool ParseStmtList(vy_parse_state *state, unsigned int *list_index);

static bool ParseStmt(vy_parse_state *state, unsigned int *stmt_index) {
    vy_parsed_stmt stmt;
    stmt.next = UINT_MAX; /* end of list */

    SkipWhitespace(state);
    if (!ParseAttribute(state, &stmt.attribute))
        return false;
    SkipWhitespace(state);

    if (state->at == state->length) {
        vyReportError("GFX",
                      "%s:%d Expected either a value or '{'",
                      state->file,
                      state->line);
        return false;
    }

    if (state->text[state->at] == '{') {
        /* Consume '{' */
        ++state->at;

        stmt.form = VY_STMT_FORM_LIST;
        if (!ParseStmtList(state, &stmt.list_index))
            return false;

        /* Consume '}' */
        if (state->at < state->length && state->text[state->at] == '}')
            ++state->at;
    } else {
        stmt.form = VY_STMT_FORM_VALUE;
        if (!ParseValue(state, &stmt.value))
            return false;

        /* Consume ';' */
        if (state->at < state->length && state->text[state->at] == ';')
            ++state->at;
    }

    SkipWhitespace(state);

    /* Add statement to array */
    if (state->statement_count == state->statement_capacity) {
        unsigned int cap = (state->statement_capacity > 0)
                               ? state->statement_capacity * 2
                               : 64;
        vy_parsed_stmt *temp =
            realloc(state->statements, sizeof(vy_parsed_stmt) * cap);
        if (!temp) {
            vyReportError("GFX",
                          "While parsing %s: Out of memory\n",
                          state->file);
            return false;
        }
        state->statements         = temp;
        state->statement_capacity = cap;
    }
    state->statements[state->statement_count] = stmt;
    *stmt_index                               = state->statement_count++;

    return true;
}

static bool ParseStmtList(vy_parse_state *state, unsigned int *list_index) {
    vy_parsed_stmt_list list;
    list.first = state->statement_count;
    list.count = 0;

    unsigned int last = UINT_MAX;

    while (state->at < state->length && state->text[state->at] != '}') {
        unsigned int stmt;
        if (!ParseStmt(state, &stmt))
            return false;

        if (last != UINT_MAX)
            state->statements[last].next = stmt;
        last = stmt;

        ++list.count;
    }

    /* Add list to array */
    if (state->statement_list_count == state->statement_list_capacity) {
        unsigned int cap = (state->statement_list_capacity > 0)
                               ? state->statement_list_capacity * 2
                               : 64;
        vy_parsed_stmt_list *temp =
            realloc(state->statement_lists, sizeof(vy_parsed_stmt_list) * cap);
        if (!temp) {
            vyReportError("GFX",
                          "While parsing %s: Out of memory\n",
                          state->file);
            return false;
        }
        state->statement_lists         = temp;
        state->statement_list_capacity = cap;
    }
    state->statement_lists[state->statement_list_count] = list;
    *list_index = state->statement_list_count++;
    return true;
}

static void DbgPrintShaderFile(const vy_parse_state *state,
                               unsigned int list_index,
                               unsigned int indent) {
    assert(list_index < state->statement_list_count);
    const vy_parsed_stmt_list *list = &state->statement_lists[list_index];

    unsigned int stmt_index = list->first;
    for (unsigned int i = 0; i < list->count; ++i) {
        const vy_parsed_stmt *stmt = &state->statements[stmt_index];
        for (unsigned int j = 0; j < indent; ++j)
            printf(" ");
        printf("%.*s: ", stmt->attribute.length, stmt->attribute.start);
        if (stmt->form == VY_STMT_FORM_VALUE) {
            printf("%.*s\n", stmt->value.length, stmt->value.start);
        } else {
            printf("{\n");
            DbgPrintShaderFile(state, stmt->list_index, indent + 2);
            printf("}\n");
        }
        stmt_index = stmt->next;
    }
    assert(stmt_index == UINT_MAX || stmt_index == 0);
}

static bool CompareSpanToString(vy_text_span span, const char *cmp) {
    size_t cmp_len = strlen(cmp);
    if (cmp_len != (size_t)span.length)
        return false;
    for (size_t i = 0; i < cmp_len; ++i) {
        if (span.start[i] != cmp[i])
            return false;
    }
    return true;
}

static const vy_parsed_stmt *FindStatement(const vy_parse_state *state,
                                           unsigned int list_index,
                                           const char *attribute) {
    if (list_index >= state->statement_list_count)
        return NULL;
    const vy_parsed_stmt_list *list = &state->statement_lists[list_index];
    unsigned int stmt_index         = list->first;
    for (unsigned int i = 0; i < list->count; ++i) {
        const vy_parsed_stmt *stmt = &state->statements[stmt_index];
        if (CompareSpanToString(stmt->attribute, attribute))
            return stmt;
        stmt_index = stmt->next;
    }
    return NULL;
}

static vy_fio_handle DispatchFileRead(vy_text_span path) {
    vy_file_id fid = vyAddFileFromSpan(path);
    if (fid == 0)
        return 0;
    vy_fio_handle fio = vyEnqueueRead(fid);
    return fio;
}

static vy_fio_handle DispatchShaderRead(const char *shader,
                                        const vy_parse_state *state,
                                        unsigned int root_list,
                                        const char *file_path) {
    const vy_parsed_stmt *path = FindStatement(state, root_list, shader);
    if (!path) {
        return 0;
    }
    if (path->form != VY_STMT_FORM_VALUE) {
        vyReportError("GFX",
                      "Expected simple value for attribute \"%s\" in %s\n",
                      shader,
                      file_path);
        return 0;
    }
    return DispatchFileRead(path->value);
}

static vy_gfx_pipeline_handle CreatePipeline(vy_parse_state *state,
                                             const char *file_path,
                                             unsigned int root_list) {
    /* Process the data */
    vy_fio_handle vertex_read =
        DispatchShaderRead("vertex", state, root_list, file_path);
    vy_fio_handle fragment_read =
        DispatchShaderRead("fragment", state, root_list, file_path);
    vy_fio_handle compute_read =
        DispatchShaderRead("compute", state, root_list, file_path);

    if (compute_read) {
        if (vertex_read || fragment_read) {
            vyReportError("GFX",
                          "A shader program can not contain a compute "
                          "shader and graphics shaders: %s",
                          file_path);
            vyAbortFIO(vertex_read);
            vyAbortFIO(fragment_read);
            vyAbortFIO(compute_read);
            return (vy_gfx_pipeline_handle){0};
        }
        while (!vyIsFIOFinished(compute_read)) {
            /* wait */
        }
        vy_file_buffer compute_code;
        if (!vyRetrieveReadBuffer(compute_read, &compute_code) ||
            !vyWasFileBufferSuccessful(&compute_code)) {
            vyReportError("GFX",
                          "Failed to load compute shader required by: %s",
                          file_path);
            return (vy_gfx_pipeline_handle){0};
        }
        vy_compute_pipeline_info info;
        info.compute_source        = compute_code.data;
        info.compute_source_length = compute_code.size;
        vy_gfx_pipeline_handle pipeline =
            g_renderer.CompileComputePipeline(&info);
        vyFreeFileBuffer(compute_code);
        return pipeline;
    } else if (vertex_read || fragment_read) {
        if (compute_read) {
            vyReportError("GFX",
                          "A shader program can not contain a compute "
                          "shader and graphics shaders: %s",
                          file_path);
            vyAbortFIO(vertex_read);
            vyAbortFIO(fragment_read);
            vyAbortFIO(compute_read);
            return (vy_gfx_pipeline_handle){0};
        }
        if (!vertex_read || !fragment_read) {
            vyReportError("GFX",
                          "A shader program must contain at least a vertex and "
                          "a fragment shader: %s",
                          file_path);
            vyAbortFIO(vertex_read);
            vyAbortFIO(fragment_read);
            vyAbortFIO(compute_read);
            return (vy_gfx_pipeline_handle){0};
        }

        vy_graphics_pipeline_info info;
        vy_file_buffer vertex_code   = {NULL, 0};
        vy_file_buffer fragment_code = {NULL, 0};

        int remaining = 2;
        while (remaining > 0) {
            if (vyIsFIOFinished(vertex_read)) {
                if (!vyRetrieveReadBuffer(vertex_read, &vertex_code) ||
                    !vyWasFileBufferSuccessful(&vertex_code)) {
                    vyReportError(
                        "GFX",
                        "Failed to load vertex shader required by: %s",
                        file_path);
                    vyFreeFileBuffer(fragment_code);
                    return (vy_gfx_pipeline_handle){0};
                }
                info.vertex_source        = vertex_code.data;
                info.vertex_source_length = vertex_code.size;
                --remaining;
                vertex_read = 0;
            }

            if (vyIsFIOFinished(fragment_read)) {
                if (!vyRetrieveReadBuffer(fragment_read, &fragment_code) ||
                    !vyWasFileBufferSuccessful(&fragment_code)) {
                    vyReportError(
                        "GFX",
                        "Failed to load fragment shader required by: %s",
                        file_path);
                    vyFreeFileBuffer(vertex_code);
                    return (vy_gfx_pipeline_handle){0};
                }
                info.fragment_source        = fragment_code.data;
                info.fragment_source_length = fragment_code.size;
                --remaining;
                fragment_read = 0;
            }
        }

        vy_gfx_pipeline_handle pipeline =
            g_renderer.CompileGraphicsPipeline(&info);
        vyFreeFileBuffer(vertex_code);
        vyFreeFileBuffer(fragment_code);
        return pipeline;
    } else if (!compute_read && !vertex_read && !fragment_read) {
        vyReportError("GFX",
                      "Either \"compute\" or \"vertex\" and \"fragment\" "
                      "are required in %s",
                      file_path);
        vyAbortFIO(vertex_read);
        vyAbortFIO(fragment_read);
        vyAbortFIO(compute_read);
    }
    return (vy_gfx_pipeline_handle){0};
}

static bool ParseBindingIndex(vy_text_span span, unsigned int *index) {
    if (span.length == 0)
        return false;
    int at           = (int)span.length - 1;
    unsigned int exp = 1;
    unsigned int n   = 0;
    while (at >= 0) {
        if (span.start[at] >= '0' && span.start[at] <= '9') {
            unsigned int digit = (unsigned int)(span.start[at] - '0');
            n += digit * exp;
        } else {
            vyReportError("GFX",
                          "Unexpected non-digit character in binding index");
            return false;
        }
        --at;
        exp *= 10;
    }
    *index = n;
    return true;
}

static vy_attribute_value ParseBindingValue(vy_text_span span) {
    if (CompareSpanToString(span, "MATERIAL_ALBEDO")) {
        return VY_ATTRIBUTE_VALUE_MATERIAL_ALBEDO;
    }
    vyReportError("GFX",
                  "Unsupported binding value %*.s",
                  span.length,
                  span.start);
    return VY_ATTRIBUTE_VALUE_UNDEFINED;
}

static bool ParseBindings(vy_parse_state *state,
                          unsigned int root_list,
                          const char *name,
                          const char *file_path,
                          vy_attribute_binding **p_bindings,
                          unsigned int *p_binding_count) {
    const vy_parsed_stmt *bindings = FindStatement(state, root_list, name);
    if (bindings) {
        if (bindings->form != VY_STMT_FORM_LIST) {
            vyReportError("GFX",
                          "Expected list of bindings as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return false;
        }
        const vy_parsed_stmt_list *binding_list =
            &state->statement_lists[bindings->list_index];
        vy_attribute_binding *shader_bindings =
            malloc(sizeof(vy_attribute_binding) * binding_list->count);
        if (!bindings) {
            vyReportError("GFX", "Out of memory");
            return false;
        }
        unsigned int binding_count = binding_list->count;

        unsigned int stmt_index = binding_list->first;
        for (unsigned int i = 0; i < binding_list->count; ++i) {
            const vy_parsed_stmt *stmt = &state->statements[stmt_index];
            if (!ParseBindingIndex(stmt->attribute,
                                   &shader_bindings[i].index)) {
                free(shader_bindings);
                return false;
            }

            shader_bindings[i].value = ParseBindingValue(stmt->value);
            if (shader_bindings[i].value == VY_ATTRIBUTE_VALUE_UNDEFINED) {
                free(shader_bindings);
                return false;
            }
            stmt_index = stmt->next;
        }

        *p_bindings      = shader_bindings;
        *p_binding_count = binding_count;
        return true;
    } else {
        *p_bindings      = NULL;
        *p_binding_count = 0;
        return true;
    }
}

static bool
ParseShaderFile(vy_file_id fid, vy_file_buffer fbuf, vy_shader *shader) {
    /* This is the grammar for shader files:
     * <stmt-list> ::= <stmt>*
     * <stmt> ::= <attribute> ( ( <value> ';' ) | ( '{' <stmt-list> '}' ) )
     * <attribute> ::= [:alnum:]*
     * <value>:: = [:alnum:]* */
    const char *file_path = vyGetFilePath(fid);
    vy_parse_state state  = {.text                    = (const char *)fbuf.data,
                             .at                      = 0,
                             .length                  = fbuf.size,
                             .line                    = 1,
                             .file                    = file_path,
                             .statements              = NULL,
                             .statement_lists         = NULL,
                             .statement_capacity      = 0,
                             .statement_list_capacity = 0};

    bool result            = true;
    unsigned int root_list = 0;
    if (!ParseStmtList(&state, &root_list)) {
        result = false;
        goto out;
    }

    DbgPrintShaderFile(&state, root_list, 0);

    shader->pipeline = CreatePipeline(&state, file_path, root_list);
    if (!VY_IS_HANDLE_VALID(shader->pipeline)) {
        result = false;
        goto out;
    }

    /* Process bindings */
    shader->texture_bindings      = NULL;
    shader->texture_binding_count = 0;
    shader->uniform_bindings      = NULL;
    shader->uniform_binding_count = 0;
    shader->storage_bindings      = NULL;
    shader->storage_binding_count = 0;

    if (!ParseBindings(&state,
                       root_list,
                       "texture_bindings",
                       file_path,
                       &shader->texture_bindings,
                       &shader->texture_binding_count)) {
        result = false;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "uniform_bindings",
                       file_path,
                       &shader->uniform_bindings,
                       &shader->uniform_binding_count)) {
        result = false;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "storage_bindings",
                       file_path,
                       &shader->storage_bindings,
                       &shader->storage_binding_count)) {
        result = false;
        goto out;
    }

out:
    free(state.statements);
    free(state.statement_lists);
    return result;
}

bool vyLoadShaders(const char **paths, vy_shader *shaders, unsigned int count) {
    vy_fio_handle fios[64];
    vy_file_id fids[64];
    for (unsigned int i = 0; i < count; i += 64) {
        unsigned int chunk_size = count - i;
        if (chunk_size > 64)
            chunk_size = 64;
        for (unsigned int j = 0; j < chunk_size; ++j) {
            vy_file_id fid = vyAddFile(paths[i + j]);
            if (!fid)
                return false;
            fids[j] = fid;
            fios[j] = vyEnqueueRead(fid);
        }

        unsigned int remaining = chunk_size;
        while (remaining > 0) {
            for (unsigned int j = 0; j < chunk_size; ++j) {
                if (vyIsFIOFinished(fios[j])) {
                    vy_file_buffer fbuf;
                    if (!vyRetrieveReadBuffer(fios[j], &fbuf)) {
                        return false;
                    }
                    ParseShaderFile(fids[j], fbuf, &shaders[i + j]);
                    vyFreeFileBuffer(fbuf);
                    --remaining;
                }
            }
        }
    }
    return true;
}
