#include "jobs.h"
#include "threading.h"

#define MAX_WORKERS    32
#define JOB_QUEUE_SIZE 2048

/* A "chunk" of iterations for a particular job */
typedef struct {
    uint32_t base_iteration;
    uint32_t iteration_count;
    vy_job_fn *fn;
    void *param;
} vy_job_chunk;

typedef struct {
    /* Queue */
    struct {
        vy_job_chunk *chunks;
        unsigned int head;
        unsigned int tail;
        vy_condition_var *lock;
    } job_queue;

    /* Thread data */
    vy_thread *thread;
} vy_worker_data;

static volatile bool _keep_running = true;
static vy_worker_data _worker_data[MAX_WORKERS];

static void ExecOneJobIfAvailable(vy_worker_data *wd) {
    if (wd->job_queue.head == wd->job_queue.tail) {
        /* No job available.
         * TODO: Pick one job queue at random and check if we can steal?
         */
        return;
    }

    vy_job_chunk chunk = wd->job_queue.chunks[wd->job_queue.head];
    wd->job_queue.head = (wd->job_queue.head + 1) % JOB_QUEUE_SIZE;
    for (uint32_t i = 0; i < chunk.iteration_count; ++i) {
        chunk.fn(chunk.param, chunk.base_iteration + i);
    }
}

static void WorkerEntry(void *param) {
    vy_worker_data *wd = param;
    while (_keep_running) {
        vyLockConditionVar(wd->job_queue.lock);
        while (wd->job_queue.head == wd->job_queue.tail && _keep_running) {
            vyWaitOnConditionVar(wd->job_queue.lock);
        }
        ExecOneJobIfAvailable(wd);
        vyUnlockConditionVar(wd->job_queue.lock, false);
    }
}

void vyInitJobSystem(unsigned int worker_count) {
    if (worker_count > MAX_WORKERS)
        worker_count = MAX_WORKERS;

    for (unsigned int i = 0; i < worker_count; ++i) {
        _worker_data[i].thread = vySpawnThread(WorkerEntry, &_worker_data[i]);
    }
}
